// -*- c-basic-offset: 4 -*-
/*
 * fullnotequeue.{cc,hh} -- queue element that notifies on full
 * Eddie Kohler
 *
 * Copyright (c) 2004 Regents of the University of California
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

#include <click/config.h>
#include "bcapnotequeue.hh"
CLICK_DECLS

BcapNoteQueue::BcapNoteQueue() : timer_(this)
{
    MOD_INC_USE_COUNT;
}

BcapNoteQueue::~BcapNoteQueue()
{
    MOD_DEC_USE_COUNT;
}

void *
BcapNoteQueue::cast(const char *n)
{
    if (strcmp(n, "BcapNoteQueue") == 0)
	return (BcapNoteQueue *)this;
    else if (strcmp(n, Notifier::NONFULL_NOTIFIER) == 0)
	return static_cast<Notifier*>(&_full_note);
    else
	return NotifierQueue::cast(n);
}

int
BcapNoteQueue::configure(Vector<String> &conf, ErrorHandler *errh)
{
    tx_ = _drops = 0; // init
    String name;

    ActiveNotifier::initialize(router()); // from notifierqueue.cc
	

    if(cp_va_parse(conf, this, errh,
			cpString, "queue name", &name,
			cpUnsigned, "queue length", &_capacity,
			cpEnd) < 0)
	return -1;

    strncpy(name_,name.c_str(), 32);
    timer_.initialize(this);
    timer_.schedule_after_ms(1000);
    _full_note.initialize(router());
    _full_note.set_signal_active(true);
    return 0;
}

void 
BcapNoteQueue::run_timer() 
{
 if(tx_ != 0 || size() != 0 || _drops != 0) 
 	click_chatter("%s: [tx = %d] [size = %d ] [drop = %d ] \n", name_, tx_ / 5 ,size() ,  _drops / 5);
 tx_ = _drops = 0; // reset
 timer_.schedule_after_ms(5000);
}


void
BcapNoteQueue::push(int, Packet *p)
{
    // Code taken from SimpleQueue::push().
    int next = next_i(_tail);

    if (next != _head) {
	_q[_tail] = p;
	_tail = next;

	int s = size();
	if (s > _highwater_length)
	    _highwater_length = s;

#if !NOTIFIERQUEUE_LOCK
	// This can leave a single packet in the queue indefinitely in
	// multithreaded Click, because of a race condition with pull().
        if (!signal_active()) 
	    wake_listeners(); 
#else
        if (s == 1) {
            _lock.acquire();
	    if (!signal_active())
	        wake_listeners();
	    _lock.release();
	}
#endif

	if (s == capacity())
	    _full_note.sleep_listeners();

    } else {
//	if (_drops == 0)
//	    click_chatter("%{element}: overflow", this);
	_drops++;
	p->kill();
    }
}

Packet *
BcapNoteQueue::pull(int)
{

    // update packets sent 
    Packet *p = deq();

    if (p) {
	_sleepiness = 0;
	if (size() == capacity() - 1)
	    _full_note.wake_listeners();
	tx_++;	
    } else if (++_sleepiness == SLEEPINESS_TRIGGER) {
#if !NOTIFIERQUEUE_LOCK
        sleep_listeners();
#else
	_lock.acquire();
	if (_head == _tail)  // if still empty...
	    sleep_listeners();
	_lock.release();
#endif
    }

    return p;
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(NotifierQueue)
EXPORT_ELEMENT(BcapNoteQueue BcapNoteQueue-BcapNoteQueue)
