#ifndef _FAKE_BIGINT_H_
#define _FAKE_BIGINT_H_

#include "fake_common.hh"
#include "gmp.h"

#ifndef CLICK_LINUXMODULE
#define u_int	unsigned int
#endif

enum { mpz_bitsperlimb = sizeof (mp_limb_t) * 8 };

#define mpz_get_rawmag mpz_get_rawmag_be
#define mpz_set_rawmag mpz_set_rawmag_be


class bigint;
void mpz_set_rawmag_le(MP_INT* mp, const char *buf, size_t size);
size_t mpz_rawsize(const MP_INT *mp);
int mpz_getbit (const MP_INT *mp, size_t bit);
void mpz_get_raw (char *buf, size_t size, const MP_INT *mp);
size_t mpz_sizeinbase2(const MP_INT*);
void mpz_set_raw(MP_INT *mp, const char *buf, size_t size);
void mpz_square(MP_INT *, const MP_INT*);
int jacobi(const bigint &a, const bigint &b);
bigint random_bigint( size_t bits);
bigint random_prime(u_int nbits, const u_int *sieve,
	const u_int sievesize, u_int iter);
bigint invert(const bigint &a, const bigint &mod);
u_long mod(const bigint &a, u_long b);
void swap(bigint &a, bigint &b);



class bigint : public MP_INT {

  public: 
	bigint() { mpz_init(this);  }
	bigint(int si) { mpz_init_set_si(this,si);  }
	bigint(char *s, int base = 0) 
		{ mpz_init_set_str(this, s, base); } 
	bigint(const bigint &b) { mpz_init_set(this, &b); } 
	~bigint() { mpz_clear(this); }

	bigint &operator= (const bigint &b) 
		{ mpz_set(this, &b);  return *this; }
	bigint &operator= (int si) 
		{ mpz_set_si(this, si);  return *this; }
	bigint &operator+= (const bigint &b) 
		{ mpz_add(this, this, &b); return *this; }
	bigint &operator-= (const bigint &b) 
		{ mpz_sub(this, this, &b);  return *this; }
	bigint &operator*= (const bigint &b) 
		{ mpz_mul(this, this, &b); return *this; }
	bigint &operator/= (const bigint &b) 
		{ mpz_tdiv_q(this, this, &b);  return *this; }
	bigint &operator%= (const bigint &b) 
		{ mpz_tdiv_r(this, this, &b); return *this; }
	bigint &operator&= (const bigint &b) 
		{ mpz_and(this, this, &b);  return *this; }
	bigint &operator^= (const bigint &b) 
		{ mpz_xor(this, this, &b);  return *this; }
	bigint &operator|= (const bigint &b) 
		{ mpz_ior(this, this, &b); return *this; }
  	bigint &operator<<= (u_long ui)
		{ mpz_mul_2exp (this, this, ui); return *this; }
  	bigint &operator>>= (u_long ui)
    		{ mpz_tdiv_q_2exp (this, this, ui); return *this; }


	inline bool operator! () { return !sgn(*this); }
	
	bool probab_prime(int reps = 25) const {
		return mpz_probab_prime_p(this, reps);
	}

	bool getbit(u_long bitno) const 
		{  return mpz_getbit(this, bitno) == 1; 
		
		 }
 
	size_t nbits() const { return mpz_sizeinbase2(this); }

	void swap(bigint *a) {
		bigint t = *a;
		*a = *this;
		*this = t;
	 }

	u_long getui() const { return mpz_get_ui(this); }
	
	inline int sgn(const bigint &a) { return mpz_sgn(&a); }

	void setraw(char * s, int len) { mpz_set_raw(this, s, len);  }

	int get_rawsize() { return mpz_rawsize(this); }

	// assumes buf is large enough to hold rawsize
  	void get_raw (char *buf, int size) const {
    		mpz_get_raw (buf, size, this);
  	};

	void setbit (u_long bitno, bool val) {
	  if(val) mpz_setbit(this, bitno);
	  else mpz_clrbit(this, bitno);
	}
};

#endif
