
#include <click/config.h>
#include "fastpass_endhost.hh"
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/router.hh>
#include <click/standard/scheduleinfo.hh>
#include <click/glue.hh>
#include <clicknet/ip.h>
#include <click/standard/alignmentinfo.hh>
#include "fastpass_common.hh"
#include "bcap_common_utils.h"
#include "token_validate.hh"
#include <click/packet_anno.hh>


FastPassEndhost::FastPassEndhost()
  : Element(2, 2), _timer(this)
{
  MOD_INC_USE_COUNT;
}

FastPassEndhost::~FastPassEndhost()
{
  MOD_DEC_USE_COUNT;
}

int
FastPassEndhost::configure(Vector<String> &conf, ErrorHandler *errh)
{
  _cksum = true;  
  String token_type_str;

  if(cp_va_parse(conf, this, errh,
	cpIPAddress, "token server", &_token_server,
	cpIPAddress, "token req source", &_token_req_source,
	cpEnd) < 0)
	return -1;

#if HAVE_FAST_CHECKSUM && FAST_CHECKSUM_ALIGNED
  // check alignment
  {
    int ans, c, o;
    ans = AlignmentInfo::query(this, 0, c, o);
    _aligned = (ans && c == 4 && o == 0);
    if (!_aligned)
      errh->warning("IP header unaligned, cannot use fast IP checksum");
    if (!ans)
      errh->message("(Try passing the configuration through `click-align'.)");
  }
#endif
  
  return 0;
}


int FastPassEndhost::initialize(ErrorHandler *errh) {
	_timer.initialize(this);
	_timer.schedule_after_ms(FP_ENDHOST_TIMER_MSEC);
	return 0;
}


void FastPassEndhost::run_timer() {
//	click_chatter("FastPass Endhost timer expired!\n");
	//TODO: automatically send refresh packets to renew capabilities
	
	_timer.schedule_after_ms(FP_ENDHOST_TIMER_MSEC);
}

void FastPassEndhost::push(int i, Packet *p) {

	if(i == OUTGOING) {
		handleOutgoing(p);
	}else if(i == INCOMING) {
		handleIncoming(p);
	}
}

void FastPassEndhost::handleIncoming(Packet *p) {
	char secflow_buf[1024];
	WritablePacket *pw = p->uniqueify();	

#ifdef CLICK_LINUXMODULE
	iphdr *ip = pw->skb()->nh.iph;
	hdr_secflow *secflowhdr = (hdr_secflow*)(ip + 1);
	unsigned char proto = ip->protocol;
	u_int ip_hdr_size  = sizeof(iphdr);
#else
	click_ip *ip = reinterpret_cast<click_ip *>(pw->data());	
	hdr_secflow* secflowhdr = (hdr_secflow*)(ip + 1);  
	unsigned char proto = ip->ip_p;
	u_int ip_hdr_size = sizeof(click_ip);
#endif

	if(proto == IPPROTO_UDP) {
		// expect all fastpass packets, except token request replies
		if(parseUDPForTokenReply(pw)) return;
	}
	
 	if(proto != IPPROTO_FANFARE){
		click_chatter("Ignoring non-fastpass packet (proto = %d)\n", 
					proto);	
		output(INCOMING).push(pw);
		return;
	}
	
	int secflow_size = sizeof(hdr_secflow); // if no token
		
	if(has_token(secflowhdr)) {
		bcap_token_hdr* token_hdr = (bcap_token_hdr*)
						(secflowhdr + 1);
		secflow_size += ((token_hdr->size & 0xff) + 4);
	}

	// copy secflow data out of the packet
  	memcpy((void*)secflow_buf, (void*)(ip + 1), secflow_size);

	proto = IPPROTO_TCP;
	if(is_udp((hdr_secflow*)secflow_buf))
		proto = IPPROTO_UDP;

#ifdef CLICK_LINUXMODULE
	iphdr *new_ip = (iphdr*)((char*)ip + secflow_size);
	memcpy((void*)new_ip, (void*)ip, sizeof(iphdr));
	short old_len = ntohs(new_ip->tot_len);
	new_ip->tot_len = htons(old_len - secflow_size);
	new_ip->protocol = proto;
	new_ip->check = 0;	
	int dest_addr = new_ip->daddr;
#else

	click_ip *new_ip = (click_ip*)((char*)ip + secflow_size);
	memcpy((void*)new_ip, (void*) ip, sizeof(click_ip));
	
	short old_len = ntohs(new_ip->ip_len);
	new_ip->ip_len = htons(old_len - secflow_size);
	new_ip->ip_p = proto;
	new_ip->ip_sum = 0;
	int dest_addr = *(int*)&new_ip->ip_dst;
#endif

        pw->set_network_header((const unsigned char*)new_ip, ip_hdr_size);
	pw->pull(secflow_size);
	unsigned short* trans_header = (unsigned short*)(new_ip + 1);
	unsigned short dest_port = ntohs(trans_header[1]); 

	//click_chatter("incoming: proto = %d dest_port = %d sec_size = %d\n",
	//		proto, dest_port & 0xFFFF, secflow_size);	

	int key = create_conn_lookup_key(dest_addr,dest_port, proto);
	flow_info* flow = flow_map.findp(key);

	if(flow == NULL) {
		flow = new flow_info();  
		flow_map.insert(key, *flow);
	}

//	print_secflow_hdr((hdr_secflow*)secflow_buf);
	requester_process_hdr(&flow->requester,(hdr_secflow*)secflow_buf);
	granter_process_hdr(&flow->granter, (hdr_secflow*)secflow_buf);  

	updateChecksum((unsigned char*)new_ip, ip_hdr_size);
	output(INCOMING).push(pw);
}

void FastPassEndhost::updateChecksum(unsigned char* ip_hdr, u_int ip_hdr_size) {

	unsigned short chk_sum = 0;
#if HAVE_FAST_CHECKSUM && FAST_CHECKSUM_ALIGNED
	if (_aligned)
 	   chk_sum = ip_fast_csum((unsigned char *)ip_hdr, ip_hdr_size >> 2);
 	else
 	   chk_sum = click_in_cksum((unsigned char *)ip_hdr, ip_hdr_size);
#elif HAVE_FAST_CHECKSUM
	chk_sum = ip_fast_csum((unsigned char *)ip_hdr, ip_hdr_size >> 2);
#else
 	chk_sum = click_in_cksum((unsigned char *)ip_hdr, ip_hdr_size);
#endif  


#ifdef CLICK_LINUXMODULE
	((iphdr*)ip_hdr)->check = chk_sum;
#else
	((click_ip*)ip_hdr)->ip_sum = chk_sum;
#endif

}

bool FastPassEndhost::parseUDPForTokenReply(Packet *p) {

	const click_ip *ip = reinterpret_cast<const click_ip *>(p->data());
	click_udp* udp = (click_udp*)(ip + 1);
	unsigned short dest_port = ntohs(udp->uh_dport); 
	if(dest_port == TOKEN_REQUEST_PORT) {
		u_int len = ntohs(udp->uh_ulen) - 
				sizeof(click_udp);
		const char *data = (const char*)(udp + 1);
		String s = String(data, len);
		token_write_handler(s,this,NULL,NULL);

		u_int dst = _token_store.get_last_loaded_dest();
		Vector<Packet*>::iterator it = _needs_token_queue.begin();
		while(it) {
			Packet *check = *it;
			const click_ip *check_ip =  reinterpret_cast<const click_ip*>
						(check->data());
			if(*(u_int*)&check_ip->ip_dst == dst){
				check->set_ip_header(check_ip, sizeof(click_ip));
				IPAddress addr = IPAddress(dst);
				// we just got a token for this packet's dest
				// send it to our input again, this time it will get a token
				click_chatter("w00t!  got back token for dest %s \n",
						addr.unparse().c_str());
				_needs_token_queue.erase(it);
				handleOutgoing(check);  
				break;
			}
			it++;
		}
		p->kill();	
		return true; 	
	}
	return false;
}

void FastPassEndhost::handleOutgoing(Packet *p) {
	char secflow_buf[1024];
	timeval now;
	click_gettimeofday(&now);
        int color = PAINT_ANNO(p);	

#ifdef CLICK_LINUXMODULE
	iphdr *old_ip = p->skb()->nh.iph;
	unsigned char proto = old_ip->protocol;
	unsigned short src_port = ((unsigned short *)(old_ip + 1))[0];
	int src_addr = old_ip->saddr;
	int dst_addr = old_ip->daddr;
#else
//	const click_ip *old_ip =  reinterpret_cast<const click_ip*>
//						(p->data());
	const click_ip *old_ip = p->ip_header();
	unsigned char proto = old_ip->ip_p;
	unsigned short src_port = ((unsigned short *)(old_ip + 1))[0]; 
	int src_addr = *(int*)&(old_ip->ip_src);
	int dst_addr = *(int*)&(old_ip->ip_dst);
#endif

	if(proto != IPPROTO_UDP && proto != IPPROTO_TCP) {
		click_chatter("outoing ignoring non tcp/udp packet with proto = %d\n", 
					proto);
		output(OUTGOING).push(p);
		return;
	}

	src_port = ntohs(src_port);
	int key = create_conn_lookup_key(src_addr, src_port, proto);
	flow_info* flow = flow_map.findp(key);

	if(flow == NULL) {
		flow = new flow_info();  
		flow_map.insert(key, *flow);
	}
	
	int secflow_size = outgoing_fill_secflow_buf(flow, (in_addr*)&dst_addr, 			secflow_buf, color);	

	if(secflow_size == TokenStore::NO_TOKEN_FOUND) {
		click_chatter("No token for packet, sending request..\n");
		// store packet and send a request
		_needs_token_queue.push_back(p);
		Packet *req = _token_requester.createRequest(dst_addr,
			_token_req_source.addr(), _token_server.addr());
		output(OUTGOING).push(req);
		return;
	}

	WritablePacket *pw = p->push(secflow_size);
#ifdef CLICK_LINUXMODULE
	iphdr *new_ip = (iphdr*)pw->data();	
	memcpy((void*) new_ip, (char*) new_ip + secflow_size,sizeof(iphdr));
	pw->set_network_header((unsigned const char*)new_ip, sizeof(iphdr));
	new_ip->protocol = IPPROTO_FANFARE;
	unsigned int ip_hdr_size = sizeof(iphdr);
	short old_len = ntohs(new_ip->tot_len);
	short new_len = old_len + secflow_size;
	new_ip->tot_len = htons(new_len);
	new_ip->check = 0;
#else
	click_ip *new_ip = reinterpret_cast<click_ip *>(pw->data());	
	memcpy((void*) new_ip, (void*)old_ip,sizeof(click_ip));
	pw->set_ip_header(new_ip, sizeof(click_ip));
 	new_ip->ip_p = IPPROTO_FANFARE;

	short old_len = ntohs(new_ip->ip_len);
	short new_len = old_len + secflow_size;
	new_ip->ip_len = htons(new_len);
	unsigned int ip_hdr_size = sizeof(click_ip);
	new_ip->ip_sum = 0;
#endif

	if(proto == IPPROTO_UDP) set_udp((hdr_secflow*)secflow_buf); 
	
	memcpy( (void*)(new_ip + 1), (void*)secflow_buf, secflow_size);
    	pw->set_timestamp_anno(now);
	
	updateChecksum((unsigned char*)new_ip, ip_hdr_size);
	output(OUTGOING).push(pw);
}

// creates lookup key for flow-map based on connection info
// see FastPass.hh for details. 
int FastPassEndhost:: create_conn_lookup_key(unsigned int local_addr,
		unsigned short local_port, unsigned char proto) {
	return (local_addr & 0xFF) << 24 | 
		(proto & 0xFF) << 16 |  
		(local_port & 0xFFFF);
}


/*
 * This function sets up the fastpass header, in the secflow_buf.
 * this copies in the correct capabilities, and attaches a token
 * if needed.  
*/

int FastPassEndhost::outgoing_fill_secflow_buf(flow_info* info, 
	struct in_addr* dest_addr, char* secflow_buf, int color)
{				      		
	int header_size = sizeof(hdr_secflow);	
	hdr_secflow * secflow = (hdr_secflow*) secflow_buf;
	
	// set-up the capability header
	memset(secflow, 0xA, sizeof(secflow));// set to all 0xA so we can visually see bad copies
	init_hdr(secflow); // zero out cap header
	granter_initialize_hdr(&info->granter, secflow);
	requester_initialize_hdr(&info->requester, secflow);

	if(color != ADDTOKEN_PAINT && color != ADD_INVALID_TOKEN_PAINT) {
		secflow->controlbits &= ~HDR_HAS_TOKEN; // toggle off
	}
	
	if(has_token(secflow)) {

	 	 // set-up the token header
	  	 bcap_token_hdr *token_hdr = (bcap_token_hdr*) 
					(secflow + 1);
	  	int token_size = 0;

	   	if(color == ADDTOKEN_PAINT) {
			// try to get a real token
			token_size = _token_store.get_token(
					*(int*)dest_addr, 
					(char*)token_hdr, true);
			click_chatter("got legit token with size = %d \n",
					token_size);
	        	if(token_size == TokenStore::NO_TOKEN_FOUND) 
				return token_size;

	   	} else if(color == ADD_INVALID_TOKEN_PAINT) {
			// get fake attacker token
			token_size = _token_store.get_bogus_token(
						(char*)token_hdr);	 
	 	}else {
			click_chatter("Added token but paint is bad\n");
		}
	  	header_size += token_size;	
	}
	return header_size;

}

int FastPassEndhost::token_write_handler
(const String &conf, Element *e, void *, ErrorHandler *errh)
{
	return ((FastPassEndhost*)e)->_token_store.add_token(conf);
}

void FastPassEndhost::add_handlers(){
	add_write_handler("add_token", token_write_handler, 0);
}

CLICK_ENDDECLS
#include <click/vector.cc>
#include <click/bighashmap.cc>
EXPORT_ELEMENT(FastPassEndhost)
ELEMENT_REQUIRES(fastpass_common)
ELEMENT_REQUIRES(bcap_granter)
ELEMENT_REQUIRES(bcap_requester)
ELEMENT_REQUIRES(token_store)

