#include <click/config.h>
#include <click/glue.hh>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/straccum.hh>
#include <clicknet/ip.h>
#include <clicknet/tcp.h>
#ifdef CLICK_LINUXMODULE
#include <click/cxxprotect.h>
#include <linux/highmem.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/ip.h>
CLICK_CXX_PROTECT 
#include <linux/sched.h>
CLICK_CXX_UNPROTECT 
#include <click/cxxunprotect.h>
#endif
CLICK_DECLS

#include <click/packet_anno.hh>
#include "fp_ttc_test_source.hh"
#include "fp_test_utils.hh"
#include "fastpass_common.hh"
#include "time_utils.h"



FPTTCTestSource::FPTTCTestSource() : Element(1,1), _timer(this)
{
	MOD_INC_USE_COUNT;
}

FPTTCTestSource::~FPTTCTestSource() 
{
	MOD_DEC_USE_COUNT;
}

int FPTTCTestSource::configure(Vector<String> &conf,
	ErrorHandler *errh) 
{
	String token_type_str;

	if(cp_va_parse(conf, this, errh,
                  cpIPAddress, "source addr", &_src,
		  cpIPAddress, "dest addr" , &_dst,
		  cpString , "token type", &token_type_str,
                  cpEnd) < 0)
		return -1;

	if(token_type_str == "tva") token_type = TOKEN_TVA;
  	else if(token_type_str == "fastpass") token_type = TOKEN_PUBKEY;
  	else {
        	click_chatter("ERROR: invalid token type string '%s' \n",
        	token_type_str.c_str());
        	return -1;
  	}

	attack_packet = make_request_packet(_src, _dst,
			 EVIL_PORT, EVIL_PORT, 20);		
	return 0;
}
  
int FPTTCTestSource::initialize(ErrorHandler *e) 
{
	req_sent = req_received = 0;
	exper_active = false;
	timer_count = 0;
	timer_interval = TIMER_INTERVAL_MSEC; 
  	_timer.initialize(this);
  	_timer.schedule_after_ms(timer_interval);


	for(int i = 10; i < 100; i = i + 10) {
		attack_info.push_back(i);	
	}

	click_chatter("Activating experiment \n");
	total_expers = 10;
	exper_num = 0;
	exper_active = true;
	init_flows();
	return 0;
}

void
FPTTCTestSource::push(int i, Packet *p)
{

	click_ip *ip = (click_ip*) (p->data());
        if(ip->ip_p != IPPROTO_TCP) {
                click_chatter("got odd proto: %d \n", ip->ip_p);
                p->kill();
                return;
        }

        click_tcp *tcp = (click_tcp*)(p->data() + sizeof(click_ip));
        u_short dst_port = ntohs(tcp->th_dport);


        if(dst_port >= NUM_LEGIT_SENDERS || dst_port < 0) {
                click_chatter("received odd dst port of %d \n", dst_port);
		p->kill();
		return;
        }      

	if(!exper_active) {
        	click_chatter("Error, received request reply when not during exper \n");
                p->kill();
                return;
	}

        // see if this could be a reply to a request that we sent
	// if so, record the ttc
        if(!flows[dst_port].reply_received) {
		
        	timeval now,diff;
        	click_gettimeofday(&now);
                timer_sub(&now, &(flows[dst_port].start), &diff);

		click_chatter("got %d : sec = %d usec = %d \n", dst_port,
			now.tv_sec, now.tv_usec);
                req_received++;
                click_chatter("Got reply id = %d  [%d of %d]) ttc = %d ! (%d )\n", 
					dst_port, req_received, 
				NUM_LEGIT_SENDERS, diff.tv_sec,
				flows[dst_port].subnetwork);
		flows[dst_port].reply_received = true;
                flows[dst_port].ttc = diff;
        }
        p->kill(); // drop packet
}



void
FPTTCTestSource::run_timer()
{

  if(!exper_active) {
        _timer.schedule_after_ms(timer_interval);
	return;
  }

  struct timeval now;
  int num_attackers = attack_info[exper_num];
 // dw: big array to hold all packet pointers we are going to send, so order can be randomized
  Packet **to_send = new Packet*[(ATTACK_PPS * num_attackers) + NUM_LEGIT_SENDERS];
  int num_packets = 0;
  click_gettimeofday(&now);

  // queue-up attack packets...
  for(int i = 0; i < ATTACK_PPS; i++) {
        for(int j = 0; j < num_attackers;j++) {
                Packet *p = attack_packet->clone();
		if(token_type == TOKEN_PUBKEY){
			SET_PAINT_ANNO(p, ADD_INVALID_TOKEN_PAINT); // tell FP-endhost
		} else {
                	SET_PAINT_ANNO(p, attacker_subnetworks[j]); // paint subnetwork
		}
                to_send[num_packets] = p;
                num_packets++;
        }
  }



  // queue up legit packets
  for(int i = 0; i < NUM_LEGIT_SENDERS; i++) {
	if(flows[i].reply_received) continue;
       
        if(less_than(&now, &(flows[i].start))){
                continue; // flow hasn't started yet
        }

        bool send_packet = false;
        if(flows[i].timeout.tv_sec == 0) {
		click_chatter("flow start: %d  %d . %d\n", flows[i].id,
			now.tv_sec, now.tv_usec);
                flows[i].start.tv_sec = now.tv_sec; // set to exact start time
           	flows[i].start.tv_usec = now.tv_usec;
		send_packet = true;
		req_sent++;
        }else if(less_than(&(flows[i].timeout), &now))       {
                send_packet = true;
        }

        if(send_packet) {
                Packet *p = flows[i].request_packet->clone();
                to_send[num_packets] = p;
                num_packets++;
                // update timeout by half a second
                flows[i].timeout = now;
                if(flows[i].timeout.tv_usec > 500000) {
                        flows[i].timeout.tv_usec -= 500000;
                        flows[i].timeout.tv_sec++;
                }else {
                        flows[i].timeout.tv_usec += 500000;
                }
        }
  }

  // cheap randimization of packet order
  for(int i = 0; i < num_packets; i++ ) {
        int r1 = random() % num_packets;
        int r2 = random() % num_packets;
        Packet *temp = to_send[r1];
        to_send[r1] = to_send[r2];
        to_send[r2] = temp;
  }

  for(int i = 0; i < num_packets; i++) {
        Packet *p = to_send[i];
        p->set_timestamp_anno(now); // send timestamp doesn't really matter
        output(0).push(p);
  }

  delete to_send;

  if(timer_count == 100) {  // every second
        timer_count = 0;
        click_chatter("req_sent = %d  req_recv = %d  \n",
                        req_sent, req_received);
  }else {
        timer_count++;
  }

  if(req_received == NUM_LEGIT_SENDERS)
	experiment_done();

  _timer.schedule_after_ms(timer_interval);

}

void FPTTCTestSource :: experiment_done() {

        click_chatter("exper:%d\n", exper_num);
        for(int i = 0; i < NUM_LEGIT_SENDERS; i++) {
                click_chatter("%d,%d \n", flows[i].ttc.tv_sec,
				flows[i].ttc.tv_usec);
        }

        if(exper_num == total_expers) {
                click_chatter("***** Finished with all experiments *****  \n");
        }else {
                exper_num++;
                init_flows();
        }
}




void
FPTTCTestSource::init_flows() {

        req_sent = req_received = 0; // reset
       
        struct timeval t;
        click_gettimeofday(&t);
	click_chatter("**** warning, no delay between experiments \n");
//        t.tv_sec += 20; // delay between expers to let router queues reach a steady state 

        // schedule legitimate flows to start at random times
        for(int i = 0; i < NUM_LEGIT_SENDERS; i++) {
                flows[i].ttc.tv_sec = 0;  // initialize ttc values to invalid
		flows[i].ttc.tv_usec = 0; 
                flows[i].id = i;
		flows[i].reply_received = false;
                flows[i].start.tv_sec = t.tv_sec + (random() % SEND_PERIOD_LEN);
                flows[i].start.tv_usec = random() % 1000000;
                flows[i].timeout.tv_sec = 0; // not sent yet
                flows[i].timeout.tv_usec = 0;
		flows[i].subnetwork = random() % NUM_SUBNETWORKS;
		Packet *p = make_request_packet(_src, _dst,i, i, 20);		
		if(token_type == TOKEN_PUBKEY){
			SET_PAINT_ANNO(p, ADDTOKEN_PAINT);
		} else {
                	SET_PAINT_ANNO(p, random() % NUM_SUBNETWORKS);
		}
        	flows[i].request_packet = p;
		t.tv_sec++;
	}


        // randomly put attackers into one of the subnetworks
        // only 1/3 of the subnetworks can have attackers,
        // use the lower half as the bad subnetworks
        int num_attackers = attack_info[exper_num];
        for(int i = 0; i < num_attackers; i++) {
                attacker_subnetworks[i] = random() % (NUM_SUBNETWORKS/2);
        }
}


void FPTTCTestSource::add_handlers(){
	add_read_handler("is_done", exper_read_handler, 0);
        add_write_handler("start_next", exper_write_handler, 0);
}

String FPTTCTestSource::exper_read_handler(Element *e, void *thunk) {
  	FPTTCTestSource *source = (FPTTCTestSource*)e;        
	if(source->req_received == NUM_LEGIT_SENDERS) return "done";
	
	return "running";
}

int FPTTCTestSource::exper_write_handler(
const String &conf, Element *e, void *, ErrorHandler *errh) {
  	FPTTCTestSource *source = (FPTTCTestSource*)e;        
        click_chatter("We have been told that the experiment is over!\n"); 
	source->experiment_done(); 
        return 0;
}



CLICK_ENDDECLS
#include <click/vector.cc>
#include <click/bighashmap.cc>
ELEMENT_REQUIRES(bcap_granter)
ELEMENT_REQUIRES(bcap_requester)
EXPORT_ELEMENT(FPTTCTestSource)
