#include <click/config.h>
#include <click/glue.hh>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/straccum.hh>
#include <click/ipaddress.hh>

#ifdef CLICK_LINUXMODULE
#include <click/cxxprotect.h>
#include <linux/highmem.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/ip.h>
#include <linux/inet.h>
CLICK_CXX_PROTECT 
#include <linux/sched.h>
CLICK_CXX_UNPROTECT 
#include <click/cxxunprotect.h>

#else

#include <iostream>
#include <fstream>
using namespace std;

#endif
CLICK_DECLS
#include "pubkey_token_handler.hh"
#include "token_validate.hh"
#include "fake_bigint.hh"
#include "rabin.hh"
#include "time_utils.h"


PubkeyTokenHandler::PubkeyTokenHandler() : Element(1,2), _timer(this),
		_duplicate_filter(FILTER_SIZE_BYTES)
{
	MOD_INC_USE_COUNT;
	num_pub_keys = 0;
}

PubkeyTokenHandler::~PubkeyTokenHandler() 
{
	MOD_DEC_USE_COUNT;
}

int PubkeyTokenHandler::configure(Vector<String> &conf,
	ErrorHandler *errh) { return 0; }
  
int PubkeyTokenHandler::initialize(ErrorHandler *e) 
{
#ifdef CLICK_USERLEVEL
  // userlevel doesn't have read/write handlers, just load from a 
  // static file (lame)

  ifstream is;
  is.open ("/users/danwent/tokens/pub.db");
  is.seekg (0, ios::end);
  int length = is.tellg();
  is.seekg (0, ios::beg);
  char *buffer = new char [length];
  is.read (buffer,length);
  String pub_key_str = String(buffer, length);
  pubkey_write_handler(pub_key_str, this, NULL, NULL);
  is.close();

  click_chatter("successfully loaded %d public keys\n", 
		num_pub_keys);
#endif
	_filter_reset_counter = 0;
	_num_valid = _num_invalid = 0; 
	_timer.initialize(this);
	_timer.schedule_after_ms(TIMER_MSEC);
	return 0; 

}

void PubkeyTokenHandler::push(int i, Packet *p) {

#ifdef CLICK_LINUXMODULE
        iphdr *ipheader = p->skb()->nh.iph;
        hdr_secflow *secflowhdr = (hdr_secflow*)(ipheader + 1);
        unsigned char proto = ipheader->protocol;
	u_int ip_dest = ipheader->daddr;
#else
        const click_ip *ip = p->ip_header();
        hdr_secflow* secflowhdr = (hdr_secflow*)(ip + 1);
        unsigned char proto = ip->ip_p;
	u_int ip_dest = *(u_int*)&(ip->ip_dst);
#endif
	// technically speaking, these are errors...
	if(proto != IPPROTO_FANFARE) {
		click_chatter("ERROR: Token handler received non-fanfare"
				" packet (proto = %d) \n", proto);
		output(0).push(p);
		return;
	}
	if(!has_token(secflowhdr)) {
		click_chatter("ERROR: Token handler received non-token packet \n");
		output(0).push(p);
		return ;
	}

	bcap_token_hdr* token_hdr = (bcap_token_hdr*) (secflowhdr + 1);
 	pubkey_token * pubkey = (pubkey_token*) (((char*)token_hdr) + 4); 

	timeval now;
	click_gettimeofday(&now);
	int diff = pubkey->expire_timestamp - now.tv_sec;
	if (diff <= 0) {
		click_chatter("The token expired %d msecs ago\n",
			-1 * diff);
		output(0).push(p);
		return;
	}

	// hash dest_ip || token identifier to detect duplicates
	char dup_test[8];
	memcpy(dup_test, (void*)&ip_dest, 4);
	memcpy(dup_test + 4, (void*)&pubkey->identifier, 4);
	
	// TODO: optimize this to a TEST-AND-SET
	// so hashes are computed only once
	if(_duplicate_filter.contains(dup_test, 8)) {
		click_chatter("id = %d is a duplicate! \n", 
			pubkey->identifier);
		output(0).push(p);
		return;
	}else {
		_duplicate_filter.insert(dup_test, 8);
	}

	bool is_valid = validate_token(pub_keys, num_pub_keys, 
					((char*)token_hdr + 4),
			token_hdr->size, ip_dest);
	if(is_valid) {
		click_chatter(
		"forwarding valid token with i = %d t = %d s = %d \n",
		pubkey->identifier, pubkey->expire_timestamp,
		pubkey->salt);		

		_num_valid++;
		if(FP_ROUTER_DEBUG) click_chatter("valid token \n");
		output(1).push(p);
	} else {
		_num_invalid++;
		if(FP_ROUTER_DEBUG) click_chatter("INVALID token \n");
		output(0).push(p);
	}	

}


void PubkeyTokenHandler::run_timer() {
	if(_num_valid != 0 || _num_invalid != 0) {
		click_chatter("valid = %d invalid = %d \n", 
			_num_valid,_num_invalid);
	}

	if(_filter_reset_counter >= BLOOM_FILTER_RESET) {
		// TODO: clear bloom filter
		_filter_reset_counter = 0;
	}else {
		_filter_reset_counter++;
	}

	_num_valid = _num_invalid = 0;
	_timer.schedule_after_ms(TIMER_MSEC);
}

void PubkeyTokenHandler::pubkey_token_test(int num_verifies) {
/*

	click_chatter("num_tokens = %d \n", num_tokens);
	if(num_tokens == 0) return;

	int num_loops = num_verifies / num_tokens;
	int total_ops = num_loops * num_tokens;

	timeval start, finish, diff;
	click_gettimeofday(&start);

	for(int i = 0; i < num_verifies; i++) {
		for(int j = 0; j < num_tokens; j++) {

			bcap_token_hdr *hdr = all_tokens[j].token_hdr;
			bool is_valid = validate_token(pub_keys, 
					num_pub_keys,
					((char*)hdr + 4), hdr->size,
				*(u_int*)&all_tokens[j].dest_ip);
			if(!is_valid) click_chatter("token %d failed \n", j);
		}

	}
	click_gettimeofday(&finish);
	timer_sub(&finish, &start, &diff);

         click_chatter("%d verifies in %d sec %d usec \n", total_ops, 
		diff.tv_sec, diff.tv_usec);

*/
}

// DW: this is just a testing function, it provides no real
// functionality
int PubkeyTokenHandler::test_token_write_handler(
const String &conf, Element *e, void *, ErrorHandler *errh) {
	// dw: ignore input and run tests
	u_int num_verifies = 4000;
	const char *s = conf.c_str();
	cp_unsigned(s,s + conf.length(), 10 , &num_verifies); 
	if(num_verifies <= 0) {
		click_chatter("invalid input for num verifies \n");
		return 1;
	}

	PubkeyTokenHandler *token_handler = (PubkeyTokenHandler*)e;
	token_handler->pubkey_token_test(num_verifies);
	return 0;
}


int PubkeyTokenHandler::pubkey_write_handler
(const String &conf, Element *e, void *, ErrorHandler *errh)
{
        Vector<String> args;
        cp_argvec(conf, args);
        for(int i = 0; i < args.size(); i++){
		((PubkeyTokenHandler*)e)->add_public_key(
			args[i].mutable_c_str());
        }
        return 0;
}
                                                                                
void PubkeyTokenHandler::add_handlers(){
        add_write_handler("add_pubkey", pubkey_write_handler, 0);
	add_write_handler("test_pubkey", test_token_write_handler, 0);
}

/*
  Adds a public key to our public key database based on the provided
  public key ASCII string.  The format is:
  <ascii ip address>:<public key in base 16 ascii>
*/
int PubkeyTokenHandler::add_public_key(char* pubkey_str) {
 
    if(num_pub_keys == MAX_NUM_PUBKEYS) {
	click_chatter("error, reach max number of keys = %d \n", 
			MAX_NUM_PUBKEYS);
	return -1; // error
    }                                                                        
    char *split = strstr(pubkey_str, ":");
    char *key = split + 1;
    *split = 0; // null terminate
                                                                                
    bigint key_int(key, 16);
    pub_keys[num_pub_keys].key = new rabin_pub(key_int);
    IPAddress addr = IPAddress(pubkey_str);
    pub_keys[num_pub_keys].dest_ip = addr.addr();
    num_pub_keys++;
    return 0; // success
}

CLICK_ENDDECLS
EXPORT_ELEMENT(PubkeyTokenHandler)
