
#include "fastpass_common.hh"
#include <click/config.h>
#include <click/confparse.hh>
#include <click/error.hh>
#include <click/router.hh>
#include <click/standard/scheduleinfo.hh>
#include <click/glue.hh>
#include <clicknet/ip.h>

#define MAX_ENDHOST_TOKENS 128
#include <click/ipaddress.hh>
#include "token_store.hh"
#include "token_validate.hh"
#include "fake_bigint.hh"


TokenStore::TokenStore() {}

TokenStore::~TokenStore() {
 HashMap<u_int, Vector < bcap_token_hdr* > >::iterator it = _token_map.begin();
 while(it) {
	u_int dest = it.key();
	Vector< bcap_token_hdr* >  *vp = _token_map.findp(dest);
	for(int i = 0; i < vp->size(); i++) {
		bcap_token_hdr *hdr = (*vp)[i];
#ifdef CLICK_LINUXMODULE
		kfree(hdr);
#else
		free(hdr);
#endif
 	}	
	_token_map.remove(dest);
 }

}

/*
*/
void TokenStore::print_all_tokens() 
{

 HashMap<u_int, Vector < bcap_token_hdr* > > ::iterator it = _token_map.begin();
 while(it) {
	u_int dest = it.key();
	Vector< bcap_token_hdr* >  *vp = _token_map.findp(dest);
	for(int i = 0; i < vp->size(); i++) {
		bcap_token_hdr *hdr = (*vp)[i];
		IPAddress addr = IPAddress(dest);
		click_chatter("Dest = %s Type = %d Size = %d \n",
			addr.unparse().c_str(), hdr->type, hdr->size);
	}
 }

}

int TokenStore::get_bogus_token(char* buf) {
	int fake_sig_len = 77;
	
	timeval now;
	click_gettimeofday(&now);
	int t = now.tv_sec + 1000000; // doesn't expire soon

	bcap_token_hdr *hdr = (bcap_token_hdr*)buf;
  	hdr->type = BTT_PublicKey;
  	hdr->size = fake_sig_len + 12;

 	pubkey_token * pubkey = (pubkey_token*) (((char*)hdr) + 
					sizeof(bcap_token_hdr)); 
	pubkey->identifier = random();
	pubkey->expire_timestamp = t;
	pubkey->salt = 0;
	memset((char*)(pubkey + 1), 0xA, fake_sig_len);
	return fake_sig_len + 16;
}

/*
  Looks for a token that can be used in sending
  a request packet to "dest_ip".  The provided buffer
  MUST be large enough.  Returns the length of the 
  actually token, 0, if no token is found.  
  The token is removed from the store if remove is true
*/
int TokenStore::get_token(u_int dest_ip, char* buf, bool remove) {
 	Vector<bcap_token_hdr* >*  vp = _token_map.findp(dest_ip);
      	if(vp == NULL || vp->size() == 0) return NO_TOKEN_FOUND;
		
	bcap_token_hdr *hdr = (*vp)[0];
	int total_len = hdr->size + 4;
      	memcpy(buf, hdr, total_len);
	// TODO: should check to make sure this has valid timestamp

	if(remove) {
		vp->erase(vp->begin()); // remove front entry
#ifdef CLICK_LINUXMODULE
		kfree(hdr);
#else
		free(hdr);
#endif
		if(vp->size() == 0)
			_token_map.remove(dest_ip);
	}

      	return total_len; 
}


int TokenStore::add_token(String tok_str) {
	char buf[512];

	Vector<String> args;
        cp_argvec(tok_str, args);
        if(args.size() != 5) {
		click_chatter("Bad number of args (%d) for token \n", 
				args.size());
		return -1;
	}
        IPAddress dst = IPAddress(args[0].c_str());
	unsigned int i, t, s;
	const char *s1 = args[1].c_str();
	cp_unsigned(s1, s1 + args[1].length(), 10, &i);
	const char *s2 = args[2].c_str();
	cp_unsigned(s2, s2 + args[2].length(), 10, &t);
	const char *s3 = args[3].c_str();
	cp_unsigned(s3, s3+ args[3].length(), 10, &s);
	bigint sig = bigint(args[4].mutable_c_str(), 16);		

	int raw_len = sig.get_rawsize();
	sig.get_raw(buf, raw_len);

	int token_len = raw_len + 12; 

#ifdef CLICK_LINUXMODULE
        bcap_token_hdr *hdr = (bcap_token_hdr*)kmalloc(token_len + 4, GFP_KERNEL);
#else
	bcap_token_hdr *hdr = (bcap_token_hdr*)malloc(token_len + 4);
#endif
  	hdr->type = BTT_PublicKey;
  	hdr->size = token_len;
	
 	pubkey_token * pubkey = (pubkey_token*) (((char*)hdr) + 
					sizeof(bcap_token_hdr)); 
	pubkey->identifier = i;
	pubkey->expire_timestamp = t;
	pubkey->salt = s;
	memcpy((char*)(pubkey + 1), buf, raw_len);

        Vector< bcap_token_hdr* >* vp = _token_map.findp(dst.addr());
	if(vp == NULL) {
		Vector< bcap_token_hdr* > v; 
		_token_map.insert(dst.addr(), v); // hash makes a copy
		vp = _token_map.findp(dst.addr());
	}
	vp->push_back(hdr);  // most recent token goes on the back..

	_last_loaded_dest = dst.addr();
	return 0;
}



CLICK_ENDDECLS
#include <click/bighashmap.cc>
#include <click/vector.cc>
ELEMENT_PROVIDES(token_store)

